'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var routes = [];

/**
 * Path to regexp 
 */
function pathToRegexp(path, keys, options) {
	/**
  * The main path matching regexp utility.
  *
  * @type {RegExp}
  */
	var PATH_REGEXP = new RegExp([
	// Match escaped characters that would otherwise appear in future matches.
	// This allows the user to escape special characters that won't transform.
	'(\\\\.)',
	// Match Express-style parameters and un-named parameters with a prefix
	// and optional suffixes. Matches appear as:
	//
	// "/:test(\\d+)?" => ["/", "test", "\d+", undefined, "?"]
	// "/route(\\d+)"  => [undefined, undefined, undefined, "\d+", undefined]
	'(?:\\:(\\w+)(?:\\(((?:\\\\.|[^\\\\()])+)\\))?|\\(((?:\\\\.|[^\\\\()])+)\\))([+*?])?'].join('|'), 'g');

	/**
  * Parse a string for the raw tokens.
  *
  * @param  {string}  str
  * @param  {Object=} options
  * @return {!Array}
  */
	function parse(str, options) {
		var tokens = [];
		var key = 0;
		var index = 0;
		var path = '';
		var defaultDelimiter = options && options.delimiter || '/';
		var delimiters = options && options.delimiters || './';
		var pathEscaped = false;
		var res;

		while ((res = PATH_REGEXP.exec(str)) !== null) {
			var m = res[0];
			var escaped = res[1];
			var offset = res.index;
			path += str.slice(index, offset);
			index = offset + m.length;

			// Ignore already escaped sequences.
			if (escaped) {
				path += escaped[1];
				pathEscaped = true;
				continue;
			}

			var prev = '';
			var next = str[index];
			var name = res[2];
			var capture = res[3];
			var group = res[4];
			var modifier = res[5];

			if (!pathEscaped && path.length) {
				var k = path.length - 1;

				if (delimiters.indexOf(path[k]) > -1) {
					prev = path[k];
					path = path.slice(0, k);
				}
			}

			// Push the current path onto the tokens.
			if (path) {
				tokens.push(path);
				path = '';
				pathEscaped = false;
			}

			var partial = prev !== '' && next !== undefined && next !== prev;
			var repeat = modifier === '+' || modifier === '*';
			var optional = modifier === '?' || modifier === '*';
			var delimiter = prev || defaultDelimiter;
			var pattern = capture || group;

			tokens.push({
				name: name || key++,
				prefix: prev,
				delimiter: delimiter,
				optional: optional,
				repeat: repeat,
				partial: partial,
				pattern: pattern ? escapeGroup(pattern) : '[^' + escapeString(delimiter) + ']+?'
			});
		}

		// Push any remaining characters.
		if (path || index < str.length) {
			tokens.push(path + str.substr(index));
		}

		return tokens;
	}

	/**
  * Compile a string to a template function for the path.
  *
  * @param  {string}             str
  * @param  {Object=}            options
  * @return {!function(Object=, Object=)}
  */
	function compile(str, options) {
		return tokensToFunction(parse(str, options));
	}

	/**
  * Expose a method for transforming tokens into the path function.
  */
	function tokensToFunction(tokens) {
		// Compile all the tokens into regexps.
		var matches = new Array(tokens.length);

		// Compile all the patterns before compilation.
		for (var i = 0; i < tokens.length; i++) {
			if (_typeof(tokens[i]) === 'object') {
				matches[i] = new RegExp('^(?:' + tokens[i].pattern + ')$');
			}
		}

		return function (data, options) {
			var path = '';
			var encode = options && options.encode || encodeURIComponent;

			for (var i = 0; i < tokens.length; i++) {
				var token = tokens[i];

				if (typeof token === 'string') {
					path += token;
					continue;
				}

				var value = data ? data[token.name] : undefined;
				var segment;

				if (Array.isArray(value)) {
					if (!token.repeat) {
						throw new TypeError('Expected "' + token.name + '" to not repeat, but got array');
					}

					if (value.length === 0) {
						if (token.optional) continue;

						throw new TypeError('Expected "' + token.name + '" to not be empty');
					}

					for (var j = 0; j < value.length; j++) {
						segment = encode(value[j]);

						if (!matches[i].test(segment)) {
							throw new TypeError('Expected all "' + token.name + '" to match "' + token.pattern + '"');
						}

						path += (j === 0 ? token.prefix : token.delimiter) + segment;
					}

					continue;
				}

				if (typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean') {
					segment = encode(String(value));

					if (!matches[i].test(segment)) {
						throw new TypeError('Expected "' + token.name + '" to match "' + token.pattern + '", but got "' + segment + '"');
					}

					path += token.prefix + segment;
					continue;
				}

				if (token.optional) {
					// Prepend partial segment prefixes.
					if (token.partial) path += token.prefix;

					continue;
				}

				throw new TypeError('Expected "' + token.name + '" to be ' + (token.repeat ? 'an array' : 'a string'));
			}

			return path;
		};
	}

	/**
  * Escape a regular expression string.
  *
  * @param  {string} str
  * @return {string}
  */
	function escapeString(str) {
		return str.replace(/([.+*?=^!:${}()[\]|/\\])/g, '\\$1');
	}

	/**
  * Escape the capturing group by escaping special characters and meaning.
  *
  * @param  {string} group
  * @return {string}
  */
	function escapeGroup(group) {
		return group.replace(/([=!:$/()])/g, '\\$1');
	}

	/**
  * Get the flags for a regexp from the options.
  *
  * @param  {Object} options
  * @return {string}
  */
	function flags(options) {
		return options && options.sensitive ? '' : 'i';
	}

	/**
  * Pull out keys from a regexp.
  *
  * @param  {!RegExp} path
  * @param  {Array=}  keys
  * @return {!RegExp}
  */
	function regexpToRegexp(path, keys) {
		if (!keys) return path;

		// Use a negative lookahead to match only capturing groups.
		var groups = path.source.match(/\((?!\?)/g);

		if (groups) {
			for (var i = 0; i < groups.length; i++) {
				keys.push({
					name: i,
					prefix: null,
					delimiter: null,
					optional: false,
					repeat: false,
					partial: false,
					pattern: null
				});
			}
		}

		return path;
	}

	/**
  * Transform an array into a regexp.
  *
  * @param  {!Array}  path
  * @param  {Array=}  keys
  * @param  {Object=} options
  * @return {!RegExp}
  */
	function arrayToRegexp(path, keys, options) {
		var parts = [];

		for (var i = 0; i < path.length; i++) {
			parts.push(pathToRegexp(path[i], keys, options).source);
		}

		return new RegExp('(?:' + parts.join('|') + ')', flags(options));
	}

	/**
  * Create a path regexp from string input.
  *
  * @param  {string}  path
  * @param  {Array=}  keys
  * @param  {Object=} options
  * @return {!RegExp}
  */
	function stringToRegexp(path, keys, options) {
		return tokensToRegExp(parse(path, options), keys, options);
	}

	/**
  * Expose a function for taking tokens and returning a RegExp.
  *
  * @param  {!Array}  tokens
  * @param  {Array=}  keys
  * @param  {Object=} options
  * @return {!RegExp}
  */
	function tokensToRegExp(tokens, keys, options) {
		options = options || {};

		var strict = options.strict;
		var end = options.end !== false;
		var delimiter = escapeString(options.delimiter || '/');
		var endsWith = [].concat(options.endsWith || []).map(escapeString).concat('$').join('|');
		var route = '';

		// Iterate over the tokens and create our regexp string.
		for (var i = 0; i < tokens.length; i++) {
			var token = tokens[i];

			if (typeof token === 'string') {
				route += escapeString(token);
			} else {
				var prefix = escapeString(token.prefix);
				var capture = '(?:' + token.pattern + ')';

				if (keys) keys.push(token);

				if (token.repeat) {
					capture += '(?:' + prefix + capture + ')*';
				}

				if (token.optional) {
					if (!token.partial) {
						capture = '(?:' + prefix + '(' + capture + '))?';
					} else {
						capture = prefix + '(' + capture + ')?';
					}
				} else {
					capture = prefix + '(' + capture + ')';
				}

				route += capture;
			}
		}

		if (end) {
			if (!strict) route += '(?:' + delimiter + ')?';

			route += endsWith === '$' ? '$' : '(?=' + endsWith + ')';
		} else {
			if (!strict) route += '(?:' + delimiter + '(?=' + endsWith + '))?';

			route += '(?=' + delimiter + '|' + endsWith + ')';
		}

		return new RegExp('^' + route, flags(options));
	}

	/**
  * Normalize the given path string, returning a regular expression.
  *
  * An empty array can be passed in for the keys, which will hold the
  * placeholder key descriptions. For example, using `/user/:id`, `keys` will
  * contain `[{ name: 'id', delimiter: '/', optional: false, repeat: false }]`.
  *
  * @param  {(string|RegExp|Array)} path
  * @param  {Array=}                keys
  * @param  {Object=}               options
  * @return {!RegExp}
  */
	function pathToRegexp(path, keys, options) {
		if (path instanceof RegExp) {
			return regexpToRegexp(path, keys);
		}

		if (Array.isArray(path)) {
			return arrayToRegexp( /** @type {!Array} */path, keys, options);
		}

		return stringToRegexp( /** @type {string} */path, keys, options);
	}

	return pathToRegexp(path, keys, options);
}

var Router = function () {
	function Router() {
		_classCallCheck(this, Router);
	}

	_createClass(Router, null, [{
		key: 'route',

		/**
   * Creates a new route
   *
   * @param {String} path
   * @param {Function} controller
   * @param {Object} config
   */
		value: function route(path, controller, config) {
			routes[path] = {
				controller: controller,
				config: config
			};
		}

		/**
   * Goes to the route
   *
   * @param {String} url
   * @param {Boolean} quiet
   */

	}, {
		key: 'go',
		value: function go(url, quiet) {
			if (quiet) {
				window.history.pushState(url, url, '#' + url);
				this.directToRoute(url, true);
			} else {
				location.hash = url;
			}
		}

		/**
   * Goes to the base directory
   */

	}, {
		key: 'goToBaseDir',
		value: function goToBaseDir() {
			var url = this.url || '/';
			var base = new String(url).substring(0, url.lastIndexOf('/'));

			this.go(base);
		}

		/**
   * Gets a query string parameter
   *
   * @param {String} name
   *
   * @returns {String} Value
   */

	}, {
		key: 'query',
		value: function query(name) {
			var url = window.location.href;

			name = name.replace(/[\[\]]/g, "\\$&");

			var regex = new RegExp("[?&]" + name + "(=([^&#]*)|&|#|$)");
			var results = regex.exec(url);

			if (!results) return null;
			if (!results[2]) return '';

			return decodeURIComponent(results[2].replace(/\+/g, " "));
		}

		/**
   * Directs to the route
   *
   * @param {String} url
   * @param {Boolean} quiet
   */

	}, {
		key: 'directToRoute',
		value: function directToRoute(url, quiet) {
			// Look for route
			var context = {};
			var route = void 0;

			// Exact match
			if (routes[url]) {
				route = routes[url];

				// Use path to regexp
			} else {
				for (var path in routes) {
					var keys = [];
					var re = pathToRegexp(path, keys);
					var values = re.exec(url);

					// A match was found
					if (re.test(url)) {
						// Set the route
						route = routes[path];

						// Add context variables (first result (0) is the entire path,
						// so assign that manually and start the counter at 1 instead)
						route.url = url;
						var counter = 1;

						var _iteratorNormalCompletion = true;
						var _didIteratorError = false;
						var _iteratorError = undefined;

						try {
							for (var _iterator = keys[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
								var key = _step.value;

								route[key.name] = values[counter];
								Router.params[key.name] = values[counter];

								counter++;
							}
						} catch (err) {
							_didIteratorError = true;
							_iteratorError = err;
						} finally {
							try {
								if (!_iteratorNormalCompletion && _iterator.return) {
									_iterator.return();
								}
							} finally {
								if (_didIteratorError) {
									throw _iteratorError;
								}
							}
						}

						break;
					}
				}
			}

			Router.url = url;

			if (route && !quiet) {
				var output = route.controller();

				if (route.config) {
					if (route.config.replaceBody) {
						document.body.innerHTML = '';

						if (output) {
							var html = output.$element || output.element || output;

							_.append(document.body, html);
						}
					}
				}
			}
		}

		/**
   * Initialise
   */

	}, {
		key: 'init',
		value: function init() {
			// Get the url
			var url = location.hash.slice(1) || '/';
			var trimmed = url.substring(0, url.indexOf('?'));

			Router.params = {};

			if (trimmed) {
				url = trimmed;
			}

			// If a check is implemented, execute it
			if (typeof Router.check === 'function') {
				Router.check(
				// Pass the proposed route
				url,

				// Cancel method
				function () {
					location.hash = Router.url;
				},

				// Proceed method
				function () {
					Router.directToRoute(url);
				});

				// If not, proceed as normal
			} else {
				Router.directToRoute(url);
			}
		}
	}]);

	return Router;
}();

window.addEventListener('hashchange', Router.init);

// Add to Crisp namespace
if (typeof Crisp === 'undefined') {
	window.Crisp = {};
}
Crisp.Router = Router;
'use strict';

/**
 * Generates a new GUID
 */

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function guid() {
    function s4() {
        return Math.floor((1 + Math.random()) * 0x10000).toString(16).substring(1);
    }

    return s4() + s4() + '-' + s4() + '-' + s4() + '-' + s4() + '-' + s4() + s4() + s4();
}

// Instances
var instances = {};

// Mutation observer
new MutationObserver(function (mutations) {
    if (!mutations) {
        return;
    }

    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
        for (var _iterator = mutations[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var mutation = _step.value;

            // If any elements were removed, clean up views
            if (mutation.removedNodes.length > 0) {
                // Wait a few cycles before removing, as the element might just have been relocated
                setTimeout(function () {
                    View.cleanup();
                });
                return;
            }
        }
    } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
            }
        } finally {
            if (_didIteratorError) {
                throw _iteratorError;
            }
        }
    }
}).observe(document.body, { childList: true, subtree: true });

/**
 * View class
 *
 * @class View
 *
 * @param {Object} params
 */

var View = function () {
    function View(params) {
        _classCallCheck(this, View);

        this.name = this.constructor.name;
        this.guid = guid();
        this.events = {};

        this._adopt(params);

        instances[this.guid] = this;

        // Placeholder element
        this.element = document.createElement('div');
        this.element.className = 'cr-placeholder';

        if (typeof jQuery !== 'undefined') {
            this.$element = $(this.element);
        }
    }

    /**
     * Backwards compatible extending
     */


    _createClass(View, [{
        key: 'getName',


        /**
         * Gets the name of this View
         */
        value: function getName() {
            var name = this.constructor.toString();
            name = name.substring('function '.length);
            name = name.substring(0, name.indexOf('('));

            return name;
        }

        /**
         * Check if view should be removed
         */

    }, {
        key: 'cleanup',
        value: function cleanup() {
            // Check if the element exists in the DOM
            var parentElement = this.element ? this.element.parentElement : null;

            while (parentElement != null) {
                if (parentElement === document.body) {
                    return;
                }

                parentElement = parentElement.parentElement;
            }

            // Check if the element is the placeholder
            if (this.element && this.element.className === 'cr-placeholder') {
                return;
            }

            // If none of the above conditions were met, remove this instance
            this.remove();
        }

        /**
         * Init
         */

    }, {
        key: '_init',
        value: function _init() {
            this.prerender();
            this._render();
            this.postrender();

            if (!this.element && this.$element && this.$element.length > 0) {
                this.element = this.$element[0];
            }

            if (!this.element) {
                return;
            }

            this.trigger('ready', this);
            this.isReady = true;
        }

        /**
         * Shorthand for .on('ready')
         */

    }, {
        key: 'ready',
        value: function ready(callback) {
            if (this.isReady) {
                callback(this);
            } else {
                this.on('ready', callback);
            }
        }

        /**
         * Adopts values
         *
         * @param {Object} values
         */

    }, {
        key: '_adopt',
        value: function _adopt(params) {
            for (var k in params) {
                this[k] = params[k];
            }

            return this;
        }

        /**
         * Runs before render
         */

    }, {
        key: 'prerender',
        value: function prerender() {}

        /**
         * Tries to update dynamic elements only, but will fall back to a full render if needed
         */

    }, {
        key: 'update',
        value: function update() {
            if (!this.element) {
                return;
            }

            var output = this._getRenderOutput();

            var oldElements = this.element.querySelectorAll('*');
            var newElements = output.querySelectorAll('*');

            if (!oldElements || !newElements) {
                return;
            }

            // A flag to check if we should do a normal render
            var shouldRender = false;

            // Checks for dynamic children
            var checkChildren = function checkChildren(oldElement, newElement) {
                if (!oldElement || !newElement) {
                    return;
                }

                // The old element is not dynamic, skip it
                if (!oldElement.dataset.crDynamicChildren) {
                    return;
                }

                // The old element is dynamic, but the new one isn't, do a normal render
                if (!newElement.dataset.crDynamicChildren) {
                    return shouldRender = true;
                }

                // Both old and new elements at the same index are dynamic,
                // and the children have changed, so replace the old element with the new
                var oldChildren = oldElement.querySelectorAll('*');
                var newChildren = newElement.querySelectorAll('*');

                if (oldChildren.length !== newChildren.length) {
                    oldElement.parentNode.replaceChild(newElement, oldElement);
                }
            };

            // Checks for dynamic content
            var checkContent = function checkContent(oldElement, newElement) {
                if (!oldElement || !newElement) {
                    return;
                }

                // The old element is not dynamic, skip it
                if (!oldElement.dataset.crDynamicContent) {
                    return;
                }

                // The old element is dynamic, but the new one isn't, do a normal render
                if (!newElement.dataset.crDynamicContent) {
                    return shouldRender = true;
                }

                // Replace old content with new
                oldElement.parentNode.replaceChild(newElement, oldElement);
            };

            // Checks for dynamic attributes
            var checkAttributes = function checkAttributes(oldElement, newElement) {
                if (!oldElement || !newElement) {
                    return;
                }

                // The old element is not dynamic, skip it
                if (!oldElement.dataset.crDynamicAttributes) {
                    return;
                }

                // The old element is dynamic, but the new one isn't, do a normal render
                if (!newElement.dataset.crDynamicAttributes) {
                    return shouldRender = true;
                }

                // Remove old attributes
                for (var i = 0; i < oldElement.attributes.length; i++) {
                    oldElement.removeAttribute(oldElement.attributes[i].name, oldElement.attributes[i].value);
                }

                // Insert new attributes
                for (var _i = 0; _i < newElement.attributes.length; _i++) {
                    oldElement.setAttribute(newElement.attributes[_i].name, newElement.attributes[_i].value);
                }
            };

            // Check root elements
            checkContent(this.element, output);
            checkChildren(this.element, output);
            checkAttributes(this.element, output);

            if (shouldRender) {
                return this._render();
            }

            // Check all child elements
            for (var i = 0; i < oldElements.length; i++) {
                checkContent(oldElements[i], newElements[i]);
                checkChildren(this.element, output);
                checkAttributes(oldElements[i], newElements[i]);

                if (shouldRender) {
                    return this._render();
                }
            }
        }

        /**
         * Gets the render output
         *
         * @returns {HTMLElement} Output
         */

    }, {
        key: '_getRenderOutput',
        value: function _getRenderOutput() {
            if (typeof this.template !== 'function') {
                return;
            }

            var output = this.template.call(this, this);

            if (!output) {
                return;
            }

            // In case of jQuery, convert to native HTMLElement
            if (typeof jQuery !== 'undefined' && output instanceof jQuery) {
                output = output[0];
            }

            return output;
        }

        /**
         * Renders this view
         */

    }, {
        key: '_render',
        value: function _render() {
            var output = this._getRenderOutput();

            if (!output) {
                return;
            }

            if (this.element.parentNode) {
                this.element.parentNode.replaceChild(output, this.element);
            }

            // jQuery
            if (typeof jQuery !== 'undefined') {
                this.$element = $(output);
                this.element = this.$element[0];

                // Native JavaScript
            } else {
                this.element = output;
            }

            this.element.dataset.guid = this.guid;
        }

        /**
         * Runs after render
         */

    }, {
        key: 'postrender',
        value: function postrender() {}

        /**
         * Removes the view from DOM and memory
         */

    }, {
        key: 'remove',
        value: function remove(timeout) {
            var _this = this;

            if (!this.destroyed) {
                this.destroyed = true;

                setTimeout(function () {
                    _this.trigger('remove');

                    // jQuery
                    if (typeof jQuery !== 'undefined' && _this.$element && _this.$element.length > 0) {
                        _this.$element.remove();

                        // Native JavaScript
                    } else if (_this.element && _this.element.parentElement) {
                        _this.element.parentElement.removeChild(_this.element);
                    }

                    delete instances[_this.guid];
                }, timeout || 0);
            }
        }

        /**
         * Call an event
         */

    }, {
        key: '_call',
        value: function _call(callback, data, ui) {
            if (typeof callback !== 'function') {
                return;
            }

            callback(data, ui, this);
        }

        /**
         * Trigger an event
         */

    }, {
        key: 'trigger',
        value: function trigger(e, obj) {
            if (this.events[e]) {
                if (typeof this.events[e] === 'function') {
                    this.events[e](obj);
                } else {
                    for (var i in this.events[e]) {
                        if (typeof this.events[e][i] === 'function') {
                            this.events[e][i](obj);
                        }
                    }
                }
            }
        }

        /**
         * Bind an event
         */

    }, {
        key: 'on',
        value: function on(e, callback) {
            var view = this;

            // No events registered, register this as the only event
            if (!this.events[e]) {
                this.events[e] = function (data) {
                    view._call(callback, data, this);
                };

                // Events have already been registered, add to callback array
            } else {
                // Only one event is registered, so convert from a single reference to an array
                if (!this.events[e].length) {
                    this.events[e] = [this.events[e]];
                }

                // Insert the event call into the array 
                this.events[e].push(function (data) {
                    view._call(callback, data, this);
                });
            }
        }

        /**
         * Check if event exists
         */

    }, {
        key: 'hasEvent',
        value: function hasEvent(name) {
            for (var k in this.events) {
                if (k == name) {
                    return true;
                }
            }

            return false;
        }

        /**
         * Fetch model data
         */

    }, {
        key: 'fetch',
        value: function fetch() {
            var _this2 = this;

            var getModel = function getModel() {
                // Get model from URL
                if (!_this2.model && typeof _this2.modelUrl === 'string') {
                    var request = new XMLHttpRequest();
                    request.open('GET', _this2.modelUrl, true);

                    request.onload = function () {
                        if (request.status >= 200 && request.status < 400) {
                            // Success!
                            var data = JSON.parse(request.responseText);

                            _this2.model = data;

                            _this2._init();
                        } else {
                            // We reached our target server, but it returned an error
                            throw new Error('Couldn\'t fetch model data');
                        }
                    };

                    request.onerror = function (e) {
                        throw e;
                    };

                    request.send();

                    // Get model with function
                } else if (!_this2.model && typeof _this2.modelFunction === 'function') {
                    _this2.modelFunction(function (data) {
                        this.model = data;

                        this._init();
                    });

                    // Just perform the initialisation
                } else {
                    _this2._init();
                }
            };

            // Get the model
            getModel();
        }

        /**
         * Gets all instances of a view type
         *
         * @param {String|View} type
         */

    }], [{
        key: 'extend',
        value: function extend(func) {
            func.prototype = Object.create(View.prototype);
            func.prototype.constructor = View;

            return func;
        }
    }, {
        key: 'getAll',
        value: function getAll(type) {
            var results = [];

            if (type) {
                for (var i in instances) {
                    var instance = instances[i];
                    var isTypeMatch = typeof type === 'function' && instance instanceof type;
                    var isStringMatch = typeof type === 'string' && instance.name === type;

                    if (isTypeMatch || isStringMatch) {
                        results.push(instance);
                    }
                }
            } else {
                results = Object.values(instances);
            }

            return results;
        }

        /**
         * Gets the first instance of a view type
         *
         * @param {String|View} type
         */

    }, {
        key: 'get',
        value: function get(type) {
            var results = View.getAll(type);
            var view = results.length > 0 ? results[0] : null;

            return view;
        }

        /**
         * Removes all views of this type
         *
         * @param {String} type
         */

    }, {
        key: 'clear',
        value: function clear(type) {
            for (var _guid in instances) {
                var instance = instances[_guid];
                var isTypeMatch = typeof type === 'function' && instance instanceof type;
                var isStringMatch = typeof type === 'string' && instance.name === type;

                if (isTypeMatch || isStringMatch) {
                    instance.remove();
                }
            }
        }

        /**
         * Cleans up orphaned views
         */

    }, {
        key: 'cleanup',
        value: function cleanup() {
            for (var _guid2 in instances) {
                instances[_guid2].cleanup();
            }
        }
    }]);

    return View;
}();

// Add to Crisp namespace


if (typeof Crisp === 'undefined') {
    window.Crisp = {};
}
Crisp.View = View;
'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var lastCondition = void 0;

/**
 * Appends content to an element
 *
 * @param {HTMLElement|SVGElement} element
 * @param {Object} content
 */
function _append(element, content) {
    if (content instanceof Array) {
        for (var _i in content) {
            _append(element, content[_i]);
        }
    } else if (content !== null && content !== undefined) {
        // Crisp view
        if (content instanceof Crisp.View) {
            _append(element, content.$element || content.element);

            // jQuery logic
        } else if (typeof jQuery !== 'undefined') {
            if (element instanceof jQuery == false) {
                element = jQuery(element);
            }

            element.append(content);

            // Native JavaScript logic
        } else {
            if (typeof content === 'number' || typeof content === 'string') {
                element.innerHTML += content.toString();
            } else {
                element.appendChild(content);
            }
        }
    }
}

/**
 * Assigns event handler shorthands to element
 * This is done to prevent extending the HTMLElement/SVGElement prototype
 *
 * @param {HTMLElement|SVGElement} element
 */
function assignEvents(element) {
    /**
     * Handles the 'addEventListener' method
     *
     * @param {String} type
     * @param {Function} callback
     */
    element.on = function on(type, callback) {
        element.addEventListener(type, callback);

        return element;
    };

    /**
     * Handles the 'removeEventListener' method
     *
     * @param {String} type
     * @param {Function} callback
     */
    element.off = function off(type, callback) {
        element.removeEventListener(type, callback);

        return element;
    };

    /**
     * Removes an element
     */
    element.remove = function remove() {
        element.parentNode.removeChild(element);
    };

    // Define shorthand methods
    var shorthands = ['blur', 'change', 'click', 'focus', 'hover', 'keydown', 'keypress', 'keyup', 'mousedown', 'mouseenter', 'mouseleave', 'mouseout', 'mouseover', 'mouseup', 'select'];

    var _loop = function _loop(shorthand) {
        element[shorthand] = function (callback) {
            return element.on(shorthand, callback);
        };
    };

    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
        for (var _iterator = shorthands[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var shorthand = _step.value;

            _loop(shorthand);
        }
    } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
            }
        } finally {
            if (_didIteratorError) {
                throw _iteratorError;
            }
        }
    }
}

/**
 * Creates an element
 *
 * @param {String} tag
 * @param {String} namespace
 * @param {Object} attributes
 * @param {Object} contents
 *
 * @returns {HTMLElement|SVGElement} element
 */
function create(tag, namespace, attributes, contents) {
    var element = !!namespace ? document.createElementNS(namespace, tag) : document.createElement(tag);

    var isContents = function isContents(obj) {
        if (obj === null || obj === undefined) {
            return false;
        }

        return obj instanceof Array || obj instanceof HTMLElement || obj instanceof SVGElement || typeof jQuery !== 'undefined' && obj instanceof jQuery || typeof obj === 'string' || typeof obj === 'number';
    };

    // The attribute parameter is the content
    if (isContents(attributes)) {
        contents = [attributes, contents];

        // The attribute parameter was defined as an object
    } else if (attributes !== undefined && attributes instanceof Object && !Array.isArray(attributes)) {
        try {
            for (var key in attributes) {
                var value = attributes[key];

                // Null, undefined or false values should not be included
                if (!value && value !== 0) {
                    continue;
                }

                // Shorthand attributes
                if (key === 'dynamicContent') {
                    key = 'data-cr-dynamic-content';
                } else if (key === 'dynamicAttributes') {
                    key = 'data-cr-dynamic-attributes';
                } else if (key === 'dynamicChildren') {
                    key = 'data-cr-dynamic-children';
                }

                element.setAttribute(key, value);
            }
        } catch (e) {
            console.log(e);
        }
    }

    _append(element, contents);

    // jQuery logic
    if (typeof jQuery !== 'undefined') {
        return jQuery(element);

        // Native JavaScript logic
    } else {
        // Assign custom event functions to element instead of extending the prototype
        assignEvents(element);

        return element;
    }
}

var Elements = function () {
    function Elements() {
        _classCallCheck(this, Elements);
    }

    _createClass(Elements, null, [{
        key: 'component',

        /**
         * Defines a new component
         *
         * @param {String} tag
         * @param {String} namespace
         * @param {Function} template
         */
        value: function component(tag, namespace, template) {
            Elements[tag] = function (model) {
                return create(tag, namespace, template(model));
            };
        }

        /**
         * Appends content using the function templating rules
         *
         * @params {HTMLElement|SVGElement} parentElement
         * @params {HTMLElement|SVGElement} contents
         */

    }, {
        key: 'append',
        value: function append(parentElement) {
            for (var _len = arguments.length, contents = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
                contents[_key - 1] = arguments[_key];
            }

            _append(parentElement, contents);
        }

        /**
         * Replaces content using the function templating rules
         *
         * @params {HTMLElement|SVGElement} parentElement
         * @params {HTMLElement|SVGElement} contents
         */

    }, {
        key: 'replace',
        value: function replace(parentElement) {
            parentElement.innerHTML = '';

            for (var _len2 = arguments.length, contents = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
                contents[_key2 - 1] = arguments[_key2];
            }

            _append(parentElement, contents);
        }

        /**
         * Encapsulates logic and renders the result
         *
         * @param {Function} func
         *
         * @param {HTMLElement|SVGElement} Contents
         */

    }, {
        key: 'do',
        value: function _do(func) {
            return func();
        }

        /**
         * Renders content based on a condition
         * 
         * @param {Boolean} condition
         * @param {HTMLElement|SVGElement} contents
         *
         * @returns {HTMLElement|SVGElement} Contents
         */

    }, {
        key: 'if',
        value: function _if(condition) {
            lastCondition = condition || false;

            if (lastCondition) {
                lastCondition = undefined;

                for (var _len3 = arguments.length, contents = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
                    contents[_key3 - 1] = arguments[_key3];
                }

                return contents;
            }
        }

        /**
         * Uses the last provided condition to simulate an "else" statement
         *
         * @param {HTMLElement|SVGElement} contents
         *
         * @returns {HTMLElement|SVGElement} Contents
         */

    }, {
        key: 'else',
        value: function _else() {
            if (typeof lastCondition === 'undefined') {
                throw new Error('No "if" statement was provided before this "else" statement');
            }

            if (!lastCondition) {
                lastCondition = undefined;

                for (var _len4 = arguments.length, contents = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
                    contents[_key4] = arguments[_key4];
                }

                return contents;
            }
        }

        /**
         * Loops through an array or object, rendering elements from model data
         *
         * @param {Object} array
         * @param {Function} callback
         *
         * @returns {HTMLElement|SVGElement} elements
         */

    }, {
        key: 'each',
        value: function each(array, callback) {
            var elements = [];

            for (var i in array) {
                var element = callback(i, array[i]);

                if (element) {
                    elements.push(element);
                }
            }

            return elements;
        }

        /**
         * Loops a given number of times, rendering elements for each pass
         *
         * @param {Number} iterations
         * @param {Function} callback
         *
         * @returns {HTMLElement|SVGElement} Elements
         */

    }, {
        key: 'loop',
        value: function loop(iterations, callback) {
            var elements = [];

            for (var i = 0; i <= iterations; i++) {
                var element = callback(i);

                if (element) {
                    elements.push(element);
                }
            }

            return elements;
        }

        /**
         * A shorthand for document.querySelector
         *
         * @param {String} query
         *
         * @returns {HTMLElement|SVGElement} Element
         */

    }, {
        key: 'find',
        value: function find(query) {
            var element = document.querySelector(query);

            if (!element) {
                return null;
            }

            if (typeof jQuery !== 'undefined') {
                return jQuery(element);
            } else {
                assignEvents(element);

                return element;
            }
        }

        /**
         * A shorthand for document.querySelectorAll
         *
         * @param {String} query
         *
         * @returns {Array} Elements
         */

    }, {
        key: 'findAll',
        value: function findAll(query) {
            var elements = document.querySelectorAll(query);

            if (!elements) {
                return [];
            }

            if (typeof jQuery !== 'undefined') {
                return jQuery(elements);
            } else {
                var array = [];

                for (var _i2 = 0; _i2 < elements.length; _i2++) {
                    array[_i2] = elements[_i2];
                    assignEvents(array[_i2]);
                }

                return array;
            }
        }
    }]);

    return Elements;
}();

// ----------
// Init all element types
// ----------


function declareMethod(tag, namespace) {
    Elements[tag] = function (attributes) {
        for (var _len5 = arguments.length, contents = Array(_len5 > 1 ? _len5 - 1 : 0), _key5 = 1; _key5 < _len5; _key5++) {
            contents[_key5 - 1] = arguments[_key5];
        }

        return create(tag, namespace, attributes, contents);
    };
}

var svgElementTypes = ['altGlyph', 'altGlyphDef', 'altGlyphItem', 'animate', 'animateColor', 'animateMotion', 'animateTransform', 'circle', 'clipPath', 'color-profile', 'cursor', 'defs', 'desc', 'discard', 'ellipse', 'feBlend', 'feColorMatrix', 'feComponentTransfer', 'feComposite', 'feConvolveMatrix', 'feDiffuseLighting', 'feDisplacementMap', 'feDistantLight', 'feDropShadow', 'feFlood', 'feFuncA', 'feFuncB', 'feFuncG', 'feFuncR', 'feGaussianBlur', 'feImage', 'feMerge', 'feMergeNode', 'feMorphology', 'feOffset', 'fePointLight', 'feSpecularLighting', 'feSpotLight', 'feTile', 'feTurbulence', 'filter', 'font', 'font-face', 'font-face-format', 'font-face-name', 'font-face-src', 'font-face-uri', 'foreignObject', 'g', 'glyph', 'glyphRef', 'hatch', 'hatchpath', 'hkern', 'image', 'line', 'linearGradient', 'marker', 'mask', 'mesh', 'meshgradient', 'meshpatch', 'meshrow', 'metadata', 'missing-glyph', 'mpath', 'path', 'pattern', 'polygon', 'polyline', 'radialGradient', 'rect', 'script', 'set', 'solidcolor', 'stop', 'svg', 'switch', 'symbol', 'text', 'textPath', 'tref', 'tspan', 'unknown', 'use', 'view', 'vkern'];

for (var i in svgElementTypes) {
    declareMethod(svgElementTypes[i], 'http://www.w3.org/2000/svg');
}

var htmlElementTypes = ['a', 'abbr', 'address', 'area', 'article', 'aside', 'audio', 'b', 'base', 'bdi', 'bdo', 'blockquote', 'body', 'br', 'button', 'canvas', 'caption', 'cite', 'code', 'col', 'colgroup', 'datalist', 'dd', 'del', 'details', 'dfn', 'dialog', 'div', 'dl', 'dt', 'em', 'embed', 'fieldset', 'figcaption', 'figure', 'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'head', 'header', 'hr', 'html', 'i', 'iframe', 'img', 'input', 'ins', 'kbd', 'keygen', 'label', 'legend', 'li', 'link', 'main', 'map', 'mark', 'menu', 'menuitem', 'meta', 'meter', 'nav', 'noscript', 'object', 'ol', 'optgroup', 'option', 'output', 'p', 'param', 'picture', 'polygon', 'pre', 'progress', 'q', 'rp', 'rt', 'ruby', 's', 'samp', 'script', 'section', 'select', 'small', 'source', 'span', 'strong', 'style', 'sub', 'summary', 'sup', 'table', 'tbody', 'td', 'textarea', 'tfoot', 'th', 'thead', 'time', 'title', 'tr', 'track', 'u', 'ul', 'var', 'video', 'wbr'];

for (var i in htmlElementTypes) {
    declareMethod(htmlElementTypes[i]);
}

// Add to Crisp namespace
if (typeof Crisp === 'undefined') {
    window.Crisp = {};
}
Crisp.Elements = Elements;
